@@ -47,6 +47,8 @@
 #include "libguile/strings.h"
 #include "libguile/vectors.h"
 #include "libguile/dynwind.h"
+#include "libguile/ports.h"
+#include "libguile/ports-internal.h"
 
 #include "libguile/validate.h"
 #include "libguile/filesys.h"
@@ -649,19 +651,15 @@ set_element (fd_set *set, SCM *ports_ready, SCM element, int pos)
       SCM_ASSERT (SCM_OPFPORTP (element), element, pos, "select");
       if (pos == SCM_ARG1)
 	{
-	  /* check whether port has buffered input.  */
-	  scm_t_port *pt = SCM_PTAB_ENTRY (element);
-      
-	  if (pt->read_pos < pt->read_end)
+	  /* Check whether port has input buffered.  */
+	  if (scm_port_buffer_can_take (SCM_PORT (element)->read_buf) > 0)
 	    use_buf = 1;
 	}
       else if (pos == SCM_ARG2)
 	{
-	  /* check whether port's output buffer has room.  */
-	  scm_t_port *pt = SCM_PTAB_ENTRY (element);
-
-	  /* > 1 since writing the last byte in the buffer causes flush.  */
-	  if (pt->write_end - pt->write_pos > 1)
+	  /* Check whether port's output buffer has room.  > 1 since
+             writing the last byte in the buffer causes flush.  */
+	  if (scm_port_buffer_can_put (SCM_PORT (element)->write_buf) > 1)
 	    use_buf = 1;
 	}
       fd = use_buf ? -1 : SCM_FPORT_FDES (element);
@@ -979,7 +977,7 @@ SCM_DEFINE (scm_fsync, "fsync", 1, 0, 0,
 
   if (SCM_OPFPORTP (object))
     {
-      scm_flush_unlocked (object);
+      scm_flush (object);
       fdes = SCM_FPORT_FDES (object);
     }
   else
@@ -994,8 +992,8 @@ SCM_DEFINE (scm_fsync, "fsync", 1, 0, 0,
 #ifdef HAVE_SYMLINK
 SCM_DEFINE (scm_symlink, "symlink", 2, 0, 0,
             (SCM oldpath, SCM newpath),
-	    "Create a symbolic link named @var{oldpath} with the value\n"
-	    "(i.e., pointing to) @var{newpath}.  The return value is\n"
+	    "Create a symbolic link named @var{newpath} with the value\n"
+	    "(i.e., pointing to) @var{oldpath}.  The return value is\n"
 	    "unspecified.")
 #define FUNC_NAME s_scm_symlink
 {
@@ -1819,12 +1817,12 @@ SCM_DEFINE (scm_closedir, "closedir", 1, 0, 0,
 static int
 scm_dir_print (SCM exp, SCM port, scm_print_state *pstate SCM_UNUSED)
 {
-  scm_puts_unlocked ("#<", port);
+  scm_puts ("#<", port);
   if (!SCM_DIR_OPEN_P (exp))
-    scm_puts_unlocked ("closed: ", port);
-  scm_puts_unlocked ("directory stream ", port);
+    scm_puts ("closed: ", port);
+  scm_puts ("directory stream ", port);
   scm_uintprint (SCM_SMOB_DATA_1 (exp), 16, port);
-  scm_putc_unlocked ('>', port);
+  scm_putc ('>', port);
   return 1;
 }
 
