@@ -313,9 +313,9 @@ SCM_DEFINE (scm_set_pointer_finalizer_x, "set-pointer-finalizer!", 2, 0, 0,
 void
 scm_i_pointer_print (SCM pointer, SCM port, scm_print_state *pstate)
 {
-  scm_puts_unlocked ("#<pointer 0x", port);
+  scm_puts ("#<pointer 0x", port);
   scm_uintprint (scm_to_uintptr_t (scm_pointer_address (pointer)), 16, port);
-  scm_putc_unlocked ('>', port);
+  scm_putc ('>', port);
 }
 
 
@@ -370,7 +370,7 @@ SCM_DEFINE (scm_string_to_pointer, "string->pointer", 1, 1, 0,
 
       ret = scm_from_pointer
         (scm_to_stringn (string, NULL, enc,
-                         scm_i_default_port_conversion_handler ()),
+                         scm_i_default_string_failed_conversion_handler ()),
          free);
 
       scm_dynwind_end ();
@@ -415,7 +415,7 @@ SCM_DEFINE (scm_pointer_to_string, "pointer->string", 1, 2, 0,
       scm_dynwind_free (enc);
 
       ret = scm_from_stringn (SCM_POINTER_VALUE (pointer), len, enc,
-                              scm_i_default_port_conversion_handler ());
+                              scm_i_default_string_failed_conversion_handler ());
 
       scm_dynwind_end ();
 
