@@ -24,13 +24,11 @@
 #endif
 
 #include <errno.h>
-#include <iconv.h>
 #include <stdio.h>
 #include <assert.h>
 
 #include <uniconv.h>
 #include <unictype.h>
-#include <c-strcase.h>
 
 #include "libguile/_scm.h"
 #include "libguile/chars.h"
@@ -41,7 +39,6 @@
 #include "libguile/macros.h"
 #include "libguile/procprop.h"
 #include "libguile/read.h"
-#include "libguile/weaks.h"
 #include "libguile/programs.h"
 #include "libguile/alist.h"
 #include "libguile/struct.h"
@@ -63,21 +60,8 @@
 
 /* Character printers.  */
 
-#define PORT_CONVERSION_HANDLER(port)		\
-  SCM_PTAB_ENTRY (port)->ilseq_handler
-
-static size_t display_string (const void *, int, size_t, SCM,
-			      scm_t_string_failed_conversion_handler);
-
-static size_t write_string (const void *, int, size_t, SCM,
-			    scm_t_string_failed_conversion_handler);
-
-static int display_character (scm_t_wchar, SCM,
-			      scm_t_string_failed_conversion_handler);
-
-static void write_character (scm_t_wchar, SCM, int);
-
-static void write_character_escaped (scm_t_wchar, int, SCM);
+static void write_string (const void *, int, size_t, SCM);
+static void write_character (scm_t_wchar, SCM);
 
 
 
@@ -343,6 +327,7 @@ quote_keywordish_symbols (void)
   (INITIAL_IDENTIFIER_MASK                                              \
    | UC_CATEGORY_MASK_Nd | UC_CATEGORY_MASK_Mc | UC_CATEGORY_MASK_Me)
 
+/* FIXME: Cache this information on the symbol, somehow.  */
 static int
 symbol_has_extended_read_syntax (SCM sym)
 {
@@ -355,30 +340,63 @@ symbol_has_extended_read_syntax (SCM sym)
 
   c = scm_i_symbol_ref (sym, 0);
 
-  /* Single dot; conflicts with dotted-pair notation.  */
-  if (len == 1 && c == '.')
-    return 1;
+  switch (c) 
+    {
+    case '\'':
+    case '`':
+    case ',':
+    case '"':
+    case ';':
+    case '#':
+      /* Some initial-character constraints.  */
+      return 1;
 
-  /* Other initial-character constraints.  */
-  if (c == '\'' || c == '`' || c == ',' || c == '"' || c == ';' || c == '#')
-    return 1;
+    case '|':
+    case '\\':
+      /* R7RS allows neither '|' nor '\' in bare symbols.  */
+      if (SCM_PRINT_R7RS_SYMBOLS_P)
+        return 1;
+      break;
+  
+    case ':':
+      /* Symbols that look like keywords.  */
+      return quote_keywordish_symbols ();
+  
+    case '.':
+      /* Single dot conflicts with dotted-pair notation.  */
+      if (len == 1)
+        return 1;
+      /* Fall through to check numbers.  */
+    case '+':
+    case '-':
+    case '0':
+    case '1':
+    case '2':
+    case '3':
+    case '4':
+    case '5':
+    case '6': 
+    case '7':
+    case '8':
+    case '9':
+     /* Number-ish symbols.  Numbers with radixes already caught be #
+        above.  */
+      if (scm_is_true (scm_i_string_to_number (scm_symbol_to_string (sym), 10)))
+        return 1;
+      break;
 
-  /* R7RS allows neither '|' nor '\' in bare symbols.  */
-  if ((c == '|' || c == '\\') && SCM_PRINT_R7RS_SYMBOLS_P)
-    return 1;
-  
-  /* Keywords can be identified by trailing colons too.  */
-  if (c == ':' || scm_i_symbol_ref (sym, len - 1) == ':')
-    return quote_keywordish_symbols ();
-  
-  /* Number-ish symbols.  */
-  if (scm_is_true (scm_i_string_to_number (scm_symbol_to_string (sym), 10)))
-    return 1;
+    default:
+      break;
+    }
   
   /* Other disallowed first characters.  */
   if (!uc_is_general_category_withtable (c, INITIAL_IDENTIFIER_MASK))
     return 1;
 
+  /* Keywords can be identified by trailing colons too.  */
+  if (scm_i_symbol_ref (sym, len - 1) == ':')
+    return quote_keywordish_symbols ();
+
   /* Otherwise, any character that's in the identifier category mask is
      fine to pass through as-is, provided it's not one of the ASCII
      delimiters like `;'.  */
@@ -400,18 +418,26 @@ symbol_has_extended_read_syntax (SCM sym)
 static void
 print_normal_symbol (SCM sym, SCM port)
 {
-  scm_display (scm_symbol_to_string (sym), port);
+  size_t len = scm_i_symbol_length (sym);
+
+  if (scm_i_is_narrow_symbol (sym))
+    {
+      const char *ptr = scm_i_symbol_chars (sym);
+      scm_c_put_latin1_chars (port, (const scm_t_uint8 *) ptr, len);
+    }
+  else
+    {
+      const scm_t_wchar *ptr = scm_i_symbol_wide_chars (sym);
+      scm_c_put_utf32_chars (port, (const scm_t_uint32 *) ptr, len);
+    }
 }
 
 static void
 print_extended_symbol (SCM sym, SCM port)
 {
   size_t pos, len;
-  scm_t_string_failed_conversion_handler strategy;
 
   len = scm_i_symbol_length (sym);
-  strategy = PORT_CONVERSION_HANDLER (port);
-
   scm_lfwrite ("#{", 2, port);
 
   for (pos = 0; pos < len; pos++)
@@ -421,13 +447,7 @@ print_extended_symbol (SCM sym, SCM port)
       if (uc_is_general_category_withtable (c,
                                             SUBSEQUENT_IDENTIFIER_MASK
                                             | UC_CATEGORY_MASK_Zs))
-        {
-          if (!display_character (c, port, strategy)
-              || (c == '\\' && !display_character (c, port, strategy)))
-            scm_encoding_error ("print_extended_symbol", errno,
-                                "cannot convert to output locale",
-                                port, SCM_MAKE_CHAR (c));
-        }
+        scm_c_put_char (port, c);
       else
         {
           scm_lfwrite ("\\x", 2, port);
@@ -443,10 +463,8 @@ static void
 print_r7rs_extended_symbol (SCM sym, SCM port)
 {
   size_t pos, len;
-  scm_t_string_failed_conversion_handler strategy;
 
   len = scm_i_symbol_length (sym);
-  strategy = PORT_CONVERSION_HANDLER (port);
 
   scm_putc ('|', port);
 
@@ -471,12 +489,7 @@ print_r7rs_extended_symbol (SCM sym, SCM port)
                                                 | UC_CATEGORY_MASK_P
                                                 | UC_CATEGORY_MASK_S)
               || (c == ' '))
-            {
-              if (!display_character (c, port, strategy))
-                scm_encoding_error ("print_r7rs_extended_symbol", errno,
-                                    "cannot convert to output locale",
-                                    port, SCM_MAKE_CHAR (c));
-            }
+            scm_c_put_char (port, c);
           else
             {
               scm_lfwrite ("\\x", 2, port);
@@ -491,8 +504,8 @@ print_r7rs_extended_symbol (SCM sym, SCM port)
 }
 
 /* FIXME: allow R6RS hex escapes instead of #{...}# or |...|.  */
-void
-scm_i_print_symbol_name (SCM sym, SCM port)
+static void
+print_symbol (SCM sym, SCM port)
 {
   if (!symbol_has_extended_read_syntax (sym))
     print_normal_symbol (sym, port);
@@ -505,8 +518,8 @@ scm_i_print_symbol_name (SCM sym, SCM port)
 void
 scm_print_symbol_name (const char *str, size_t len, SCM port)
 {
-  SCM symbol = scm_from_locale_symboln (str, len);
-  scm_i_print_symbol_name (symbol, port);
+  SCM symbol = scm_from_utf8_symboln (str, len);
+  print_symbol (symbol, port);
 }
 
 /* Print generally.  Handles both write and display according to PSTATE.
@@ -517,21 +530,6 @@ SCM_GPROC(s_display, "display", 1, 1, 0, scm_display, g_display);
 static void iprin1 (SCM exp, SCM port, scm_print_state *pstate);
 
 
-/* Print a character as an octal or hex escape.  */
-#define PRINT_CHAR_ESCAPE(i, port)              \
-  do                                            \
-    {                                           \
-      if (!SCM_R6RS_ESCAPES_P)                  \
-        scm_intprint (i, 8, port);              \
-      else                                      \
-        {                                       \
-          scm_puts ("x", port);                 \
-          scm_intprint (i, 16, port);           \
-        }                                       \
-    }                                           \
-  while (0)
-
-  
 void 
 scm_iprin1 (SCM exp, SCM port, scm_print_state *pstate)
 {
@@ -546,6 +544,33 @@ scm_iprin1 (SCM exp, SCM port, scm_print_state *pstate)
     iprin1 (exp, port, pstate);
 }
 
+static void
+print_vector_or_weak_vector (SCM v, size_t len, SCM (*ref) (SCM, size_t),
+                             SCM port, scm_print_state *pstate)
+{
+  long i;
+  long last = len - 1;
+  int cutp = 0;
+  if (pstate->fancyp && len > pstate->length)
+    {
+      last = pstate->length - 1;
+      cutp = 1;
+    }
+  for (i = 0; i < last; ++i)
+    {
+      scm_iprin1 (ref (v, i), port, pstate);
+      scm_putc (' ', port);
+    }
+  if (i == last)
+    {
+      /* CHECK_INTS; */
+      scm_iprin1 (ref (v, i), port, pstate);
+    }
+  if (cutp)
+    scm_puts (" ...", port);
+  scm_putc (')', port);
+}
+
 static void
 iprin1 (SCM exp, SCM port, scm_print_state *pstate)
 {
@@ -567,15 +592,9 @@ iprin1 (SCM exp, SCM port, scm_print_state *pstate)
       if (SCM_CHARP (exp))
 	{
 	  if (SCM_WRITINGP (pstate))
-	    write_character (SCM_CHAR (exp), port, 0);
+	    write_character (SCM_CHAR (exp), port);
 	  else
-	    {
-	      if (!display_character (SCM_CHAR (exp), port,
-				      PORT_CONVERSION_HANDLER (port)))
-		scm_encoding_error (__func__, errno,
-				    "cannot convert to output locale",
-				    port, exp);
-	    }
+            scm_c_put_char (port, SCM_CHAR (exp));
 	}
       else if (SCM_IFLAGP (exp)
 	       && ((size_t) SCM_IFLAGNUM (exp) < (sizeof iflagnames / sizeof (char *))))
@@ -601,7 +620,7 @@ iprin1 (SCM exp, SCM port, scm_print_state *pstate)
 		  goto print_struct;
 		pwps = scm_i_port_with_print_state (port, pstate->handle);
 		pstate->revealed = 1;
-		scm_call_generic_2 (print, exp, pwps);
+		scm_call_2 (print, exp, pwps);
 	      }
 	    else
 	      {
@@ -636,43 +655,32 @@ iprin1 (SCM exp, SCM port, scm_print_state *pstate)
             break;
           }
 	  break;
+        case scm_tc7_stringbuf:
+          scm_i_print_stringbuf (exp, port, pstate);
+          break;
         case scm_tc7_string:
 	  {
-	    size_t len, printed;
+	    size_t len = scm_i_string_length (exp);
 
-	    len = scm_i_string_length (exp);
 	    if (SCM_WRITINGP (pstate))
-	      {
-		printed = write_string (scm_i_string_data (exp),
-					scm_i_is_narrow_string (exp),
-					len, port,
-					PORT_CONVERSION_HANDLER (port));
-		len += 2;		    /* account for the quotes */
-	      }
+              write_string (scm_i_string_data (exp),
+                            scm_i_is_narrow_string (exp),
+                            len, port);
 	    else
-	      printed = display_string (scm_i_string_data (exp),
-					scm_i_is_narrow_string (exp),
-					len, port,
-					PORT_CONVERSION_HANDLER (port));
-
-	    if (SCM_UNLIKELY (printed < len))
-	      scm_encoding_error (__func__, errno,
-				  "cannot convert to output locale",
-				  port, scm_c_string_ref (exp, printed));
+              scm_c_put_string (port, exp, 0, len);
 	  }
-
           scm_remember_upto_here_1 (exp);
           break;
 	case scm_tc7_symbol:
 	  if (scm_i_symbol_is_interned (exp))
 	    {
-	      scm_i_print_symbol_name (exp, port);
+	      print_symbol (exp, port);
 	      scm_remember_upto_here_1 (exp);
 	    }
 	  else
 	    {
 	      scm_puts ("#<uninterned-symbol ", port);
-	      scm_i_print_symbol_name (exp, port);
+	      print_symbol (exp, port);
 	      scm_putc (' ', port);
 	      scm_uintprint (SCM_UNPACK (exp), 16, port);
 	      scm_putc ('>', port);
@@ -690,6 +698,12 @@ iprin1 (SCM exp, SCM port, scm_print_state *pstate)
 	case scm_tc7_hashtable:
 	  scm_i_hashtable_print (exp, port, pstate);
 	  break;
+	case scm_tc7_weak_set:
+	  scm_i_weak_set_print (exp, port, pstate);
+	  break;
+	case scm_tc7_weak_table:
+	  scm_i_weak_table_print (exp, port, pstate);
+	  break;
 	case scm_tc7_fluid:
 	  scm_i_fluid_print (exp, port, pstate);
 	  break;
@@ -699,21 +713,13 @@ iprin1 (SCM exp, SCM port, scm_print_state *pstate)
 	case scm_tc7_frame:
 	  scm_i_frame_print (exp, port, pstate);
 	  break;
-	case scm_tc7_objcode:
-	  scm_i_objcode_print (exp, port, pstate);
-	  break;
-	case scm_tc7_vm:
-	  scm_i_vm_print (exp, port, pstate);
-	  break;
+        case scm_tc7_keyword:
+          scm_puts ("#:", port);
+          scm_iprin1 (scm_keyword_to_symbol (exp), port, pstate);
+          break;
 	case scm_tc7_vm_cont:
 	  scm_i_vm_cont_print (exp, port, pstate);
 	  break;
-	case scm_tc7_prompt:
-	  scm_i_prompt_print (exp, port, pstate);
-	  break;
-	case scm_tc7_with_fluids:
-	  scm_i_with_fluids_print (exp, port, pstate);
-	  break;
 	case scm_tc7_array:
 	  ENTER_NESTED_DATA (pstate, exp, circref);
           scm_i_print_array (exp, port, pstate);
@@ -727,65 +733,22 @@ iprin1 (SCM exp, SCM port, scm_print_state *pstate)
 	  break;
 	case scm_tc7_wvect:
 	  ENTER_NESTED_DATA (pstate, exp, circref);
-	  if (SCM_IS_WHVEC (exp))
-	    scm_puts ("#wh(", port);
-	  else
-	    scm_puts ("#w(", port);
-	  goto common_vector_printer;
+          scm_puts ("#w(", port);
+          print_vector_or_weak_vector (exp, scm_c_weak_vector_length (exp),
+                                       scm_c_weak_vector_ref, port, pstate);
+	  EXIT_NESTED_DATA (pstate);
+	  break;
 	case scm_tc7_vector:
 	  ENTER_NESTED_DATA (pstate, exp, circref);
 	  scm_puts ("#(", port);
-	common_vector_printer:
-	  {
-	    register long i;
-	    long last = SCM_SIMPLE_VECTOR_LENGTH (exp) - 1;
-	    int cutp = 0;
-	    if (pstate->fancyp
-		&& SCM_SIMPLE_VECTOR_LENGTH (exp) > pstate->length)
-	      {
-		last = pstate->length - 1;
-		cutp = 1;
-	      }
-	    if (SCM_I_WVECTP (exp))
-	      {
-		/* Elements of weak vectors may not be accessed via the
-		   `SIMPLE_VECTOR_REF ()' macro.  */
-		for (i = 0; i < last; ++i)
-		  {
-		    scm_iprin1 (scm_c_weak_vector_ref (exp, i),
-				port, pstate);
-		    scm_putc (' ', port);
-		  }
-	      }
-	    else
-	      {
-		for (i = 0; i < last; ++i)
-		  {
-		    scm_iprin1 (SCM_SIMPLE_VECTOR_REF (exp, i), port, pstate);
-		    scm_putc (' ', port);
-		  }
-	      }
-
-	    if (i == last)
-	      {
-		/* CHECK_INTS; */
-                scm_iprin1 (SCM_I_WVECTP (exp)
-                            ? scm_c_weak_vector_ref (exp, i)
-                            : SCM_SIMPLE_VECTOR_REF (exp, i),
-                            port, pstate);
-	      }
-	    if (cutp)
-	      scm_puts (" ...", port);
-	    scm_putc (')', port);
-	  }
+          print_vector_or_weak_vector (exp, SCM_SIMPLE_VECTOR_LENGTH (exp),
+                                       scm_c_vector_ref, port, pstate);
 	  EXIT_NESTED_DATA (pstate);
 	  break;
 	case scm_tc7_port:
 	  {
-	    register long i = SCM_PTOBNUM (exp);
-	    if (i < scm_numptob
-		&& scm_ptobs[i].print
-		&& (scm_ptobs[i].print) (exp, port, pstate))
+	    scm_t_port_type *ptob = SCM_PORT_TYPE (exp);
+	    if (ptob->print && ptob->print (exp, port, pstate))
 	      break;
 	    goto punk;
 	  }
@@ -860,438 +823,89 @@ scm_prin1 (SCM exp, SCM port, int writingp)
     }
 }
 
-/* Convert codepoint CH to UTF-8 and store the result in UTF8.  Return
-   the number of bytes of the UTF-8-encoded string.  */
-static size_t
-codepoint_to_utf8 (scm_t_wchar ch, scm_t_uint8 utf8[4])
-{
-  size_t len;
-  scm_t_uint32 codepoint;
-
-  codepoint = (scm_t_uint32) ch;
-
-  if (codepoint <= 0x7f)
-    {
-      len = 1;
-      utf8[0] = (scm_t_uint8) codepoint;
-    }
-  else if (codepoint <= 0x7ffUL)
-    {
-      len = 2;
-      utf8[0] = 0xc0 | (codepoint >> 6);
-      utf8[1] = 0x80 | (codepoint & 0x3f);
-    }
-  else if (codepoint <= 0xffffUL)
-    {
-      len = 3;
-      utf8[0] = 0xe0 | (codepoint >> 12);
-      utf8[1] = 0x80 | ((codepoint >> 6) & 0x3f);
-      utf8[2] = 0x80 | (codepoint & 0x3f);
-    }
-  else
-    {
-      len = 4;
-      utf8[0] = 0xf0 | (codepoint >> 18);
-      utf8[1] = 0x80 | ((codepoint >> 12) & 0x3f);
-      utf8[2] = 0x80 | ((codepoint >> 6) & 0x3f);
-      utf8[3] = 0x80 | (codepoint & 0x3f);
-    }
-
-  return len;
-}
-
-#define STR_REF(s, x)				\
-  (narrow_p					\
-   ? (scm_t_wchar) ((unsigned char *) (s))[x]	\
-   : ((scm_t_wchar *) (s))[x])
-
-/* Write STR to PORT as UTF-8.  STR is a LEN-codepoint string; it is
-   narrow if NARROW_P is true, wide otherwise.  Return LEN.  */
-static size_t
-display_string_as_utf8 (const void *str, int narrow_p, size_t len,
-			SCM port)
-{
-  size_t printed = 0;
-
-  while (len > printed)
-    {
-      size_t utf8_len, i;
-      char *input, utf8_buf[256];
-
-      /* Convert STR to UTF-8.  */
-      for (i = printed, utf8_len = 0, input = utf8_buf;
-	   i < len && utf8_len + 4 < sizeof (utf8_buf);
-	   i++)
-	{
-	  utf8_len += codepoint_to_utf8 (STR_REF (str, i),
-					 (scm_t_uint8 *) input);
-	  input = utf8_buf + utf8_len;
-	}
-
-      /* INPUT was successfully converted, entirely; print the
-	 result.  */
-      scm_lfwrite (utf8_buf, utf8_len, port);
-      printed += i - printed;
-    }
-
-  assert (printed == len);
-
-  return len;
-}
-
-/* Convert STR through PORT's output conversion descriptor and write the
-   output to PORT.  Return the number of codepoints written.  */
-static size_t
-display_string_using_iconv (const void *str, int narrow_p, size_t len,
-			    SCM port,
-			    scm_t_string_failed_conversion_handler strategy)
-{
-  size_t printed;
-  scm_t_iconv_descriptors *id;
-  scm_t_port_internal *pti = SCM_PORT_GET_INTERNAL (port);
-
-  id = scm_i_port_iconv_descriptors (port, SCM_PORT_WRITE);
-
-  if (SCM_UNLIKELY (pti->at_stream_start_for_bom_write && len > 0))
-    {
-      scm_t_port *pt = SCM_PTAB_ENTRY (port);
-
-      /* Record that we're no longer at stream start.  */
-      pti->at_stream_start_for_bom_write = 0;
-      if (pt->rw_random)
-        pti->at_stream_start_for_bom_read = 0;
-
-      /* Write a BOM if appropriate.  */
-      if (SCM_UNLIKELY (c_strcasecmp(pt->encoding, "UTF-16") == 0
-                        || c_strcasecmp(pt->encoding, "UTF-32") == 0))
-        display_character (SCM_UNICODE_BOM, port, iconveh_error);
-    }
-
-  printed = 0;
-
-  while (len > printed)
-    {
-      size_t done, utf8_len, input_left, output_left, i;
-      size_t codepoints_read, output_len;
-      char *input, *output;
-      char utf8_buf[256], encoded_output[256];
-      size_t offsets[256];
-
-      /* Convert STR to UTF-8.  */
-      for (i = printed, utf8_len = 0, input = utf8_buf;
-	   i < len && utf8_len + 4 < sizeof (utf8_buf);
-	   i++)
-	{
-	  offsets[utf8_len] = i;
-	  utf8_len += codepoint_to_utf8 (STR_REF (str, i),
-					 (scm_t_uint8 *) input);
-	  input = utf8_buf + utf8_len;
-	}
-
-      input = utf8_buf;
-      input_left = utf8_len;
-
-      output = encoded_output;
-      output_left = sizeof (encoded_output);
-
-      done = iconv (id->output_cd, &input, &input_left,
-		    &output, &output_left);
-
-      output_len = sizeof (encoded_output) - output_left;
-
-      if (SCM_UNLIKELY (done == (size_t) -1))
-	{
-          int errno_save = errno;
-
-	  /* Reset the `iconv' state.  */
-	  iconv (id->output_cd, NULL, NULL, NULL, NULL);
-
-	  /* Print the OUTPUT_LEN bytes successfully converted.  */
-	  scm_lfwrite (encoded_output, output_len, port);
-
-	  /* See how many input codepoints these OUTPUT_LEN bytes
-	     corresponds to.  */
-	  codepoints_read = offsets[input - utf8_buf] - printed;
-	  printed += codepoints_read;
-
-	  if (errno_save == EILSEQ &&
-	      strategy != SCM_FAILED_CONVERSION_ERROR)
-	    {
-	      /* Conversion failed somewhere in INPUT and we want to
-		 escape or substitute the offending input character.  */
-
-	      if (strategy == SCM_FAILED_CONVERSION_ESCAPE_SEQUENCE)
-		{
-		  scm_t_wchar ch;
-
-		  /* Find CH, the offending codepoint, and escape it.  */
-		  ch = STR_REF (str, offsets[input - utf8_buf]);
-		  write_character_escaped (ch, 1, port);
-		}
-	      else
-		/* STRATEGY is `SCM_FAILED_CONVERSION_QUESTION_MARK'.  */
-		display_string ("?", 1, 1, port, strategy);
-
-	      printed++;
-	    }
-	  else
-	    /* Something bad happened that we can't handle: bail out.  */
-	    break;
-	}
-      else
-	{
-	  /* INPUT was successfully converted, entirely; print the
-	     result.  */
-	  scm_lfwrite (encoded_output, output_len, port);
-	  codepoints_read = i - printed;
-	  printed += codepoints_read;
-	}
-    }
-
-  return printed;
-}
-
-/* Display the LEN codepoints in STR to PORT according to STRATEGY;
-   return the number of codepoints successfully displayed.  If NARROW_P,
-   then STR is interpreted as a sequence of `char', denoting a Latin-1
-   string; otherwise it's interpreted as a sequence of
-   `scm_t_wchar'.  */
-static size_t
-display_string (const void *str, int narrow_p,
-		size_t len, SCM port,
-		scm_t_string_failed_conversion_handler strategy)
-
-{
-  scm_t_port_internal *pti;
-
-  pti = SCM_PORT_GET_INTERNAL (port);
-
-  if (pti->encoding_mode == SCM_PORT_ENCODING_MODE_UTF8)
-    return display_string_as_utf8 (str, narrow_p, len, port);
-  else
-    return display_string_using_iconv (str, narrow_p, len,
-				       port, strategy);
-}
-
-/* Attempt to display CH to PORT according to STRATEGY.  Return one if
-   CH was successfully displayed, zero otherwise (e.g., if it was not
-   representable in PORT's encoding.)  */
-static int
-display_character (scm_t_wchar ch, SCM port,
-		   scm_t_string_failed_conversion_handler strategy)
-{
-  return display_string (&ch, 0, 1, port, strategy) == 1;
-}
-
-/* Same as 'display_string', but using the 'write' syntax.  */
-static size_t
-write_string (const void *str, int narrow_p,
-	      size_t len, SCM port,
-	      scm_t_string_failed_conversion_handler strategy)
-{
-  size_t printed;
-
-  printed = display_character ('"', port, strategy);
-
-  if (printed > 0)
-    {
-      size_t i;
-
-      for (i = 0; i < len; ++i)
-	{
-	  write_character (STR_REF (str, i), port, 1);
-	  printed++;
-	}
-
-      printed += display_character ('"', port, strategy);
-    }
-
-  return printed;
-}
-
-#undef STR_REF
-
-/* Attempt to pretty-print CH, a combining character, to PORT.  Return
-   zero upon failure, non-zero otherwise.  The idea is to print CH above
-   a dotted circle to make it more visible.  */
-static int
-write_combining_character (scm_t_wchar ch, SCM port)
-{
-  scm_t_wchar str[2];
-
-  str[0] = SCM_CODEPOINT_DOTTED_CIRCLE;
-  str[1] = ch;
-
-  return display_string (str, 0, 2, port, iconveh_error) == 2;
-}
-
-/* Write CH to PORT in its escaped form, using the string escape syntax
-   if STRING_ESCAPES_P is non-zero.  */
 static void
-write_character_escaped (scm_t_wchar ch, int string_escapes_p, SCM port)
+write_string (const void *str, int narrow_p, size_t len, SCM port)
 {
-  if (string_escapes_p)
+  size_t i;
+
+  scm_c_put_char (port, (scm_t_uint8) '"');
+
+  for (i = 0; i < len; ++i)
     {
-      /* Represent CH using the in-string escape syntax.  */
-
-      static const char hex[] = "0123456789abcdef";
-      static const char escapes[7] = "abtnvfr";
-      char buf[9];
-
-      if (ch >= 0x07 && ch <= 0x0D && ch != 0x0A)
-	{
-	  /* Use special escapes for some C0 controls.  */
-	  buf[0] = '\\';
-	  buf[1] = escapes[ch - 0x07];
-	  scm_lfwrite (buf, 2, port);
-	}
-      else if (!SCM_R6RS_ESCAPES_P)
-	{
-	  if (ch <= 0xFF)
-	    {
-	      buf[0] = '\\';
-	      buf[1] = 'x';
-	      buf[2] = hex[ch / 16];
-	      buf[3] = hex[ch % 16];
-	      scm_lfwrite (buf, 4, port);
-	    }
-	  else if (ch <= 0xFFFF)
-	    {
-	      buf[0] = '\\';
-	      buf[1] = 'u';
-	      buf[2] = hex[(ch & 0xF000) >> 12];
-	      buf[3] = hex[(ch & 0xF00) >> 8];
-	      buf[4] = hex[(ch & 0xF0) >> 4];
-	      buf[5] = hex[(ch & 0xF)];
-	      scm_lfwrite (buf, 6, port);
-	    }
-	  else if (ch > 0xFFFF)
-	    {
-	      buf[0] = '\\';
-	      buf[1] = 'U';
-	      buf[2] = hex[(ch & 0xF00000) >> 20];
-	      buf[3] = hex[(ch & 0xF0000) >> 16];
-	      buf[4] = hex[(ch & 0xF000) >> 12];
-	      buf[5] = hex[(ch & 0xF00) >> 8];
-	      buf[6] = hex[(ch & 0xF0) >> 4];
-	      buf[7] = hex[(ch & 0xF)];
-	      scm_lfwrite (buf, 8, port);
-	    }
-	}
+      scm_t_wchar ch;
+      if (narrow_p)
+        ch = (scm_t_wchar) ((unsigned char *) (str))[i];
       else
-	{
-	  /* Print an R6RS variable-length hex escape: "\xNNNN;".  */
-	  scm_t_wchar ch2 = ch;
+        ch = ((scm_t_wchar *) (str))[i];
 
-	  int i = 8;
-	  buf[i] = ';';
-	  i --;
-	  if (ch == 0)
-	    buf[i--] = '0';
-	  else
-	    while (ch2 > 0)
-	      {
-		buf[i] = hex[ch2 & 0xF];
-		ch2 >>= 4;
-		i --;
-	      }
-	  buf[i] = 'x';
-	  i --;
-	  buf[i] = '\\';
-	  scm_lfwrite (buf + i, 9 - i, port);
-	}
+      /* Write CH to PORT, escaping it if it's non-graphic or not
+         representable in PORT's encoding.  If CH needs to be escaped,
+         it is escaped using the in-string escape syntax.  */
+      if (ch == '"')
+        scm_c_put_latin1_chars (port, (const scm_t_uint8 *) "\\\"", 2);
+      else if (ch == '\\')
+        scm_c_put_latin1_chars (port, (const scm_t_uint8 *) "\\\\", 2);
+      else if (ch == '\n' && SCM_PRINT_ESCAPE_NEWLINES_P)
+        scm_c_put_latin1_chars (port, (const scm_t_uint8 *) "\\n", 2);
+      else if (ch == ' ' || ch == '\n'
+               || (uc_is_general_category_withtable (ch,
+                                                     UC_CATEGORY_MASK_L |
+                                                     UC_CATEGORY_MASK_M |
+                                                     UC_CATEGORY_MASK_N |
+                                                     UC_CATEGORY_MASK_P |
+                                                     UC_CATEGORY_MASK_S)
+                   && scm_c_can_put_char (port, ch)))
+        scm_c_put_char (port, ch);
+      else
+        scm_c_put_escaped_char (port, ch);
     }
+
+  scm_c_put_char (port, (scm_t_uint8) '"');
+}
+
+/* Write CH to PORT, escaping it if it's non-graphic or not
+   representable in PORT's encoding.  The character escape syntax is
+   used.  */
+static void
+write_character (scm_t_wchar ch, SCM port)
+{
+  scm_puts ("#\\", port);
+
+  /* Pretty-print a combining characters over dotted circles, if
+     possible, to make them more visible.  */
+  if (uc_combining_class (ch) != UC_CCC_NR
+      && scm_c_can_put_char (port, SCM_CODEPOINT_DOTTED_CIRCLE)
+      && scm_c_can_put_char (port, ch))
+    {
+      scm_c_put_char (port, SCM_CODEPOINT_DOTTED_CIRCLE);
+      scm_c_put_char (port, ch);
+    }
+  else if (uc_is_general_category_withtable (ch,
+                                             UC_CATEGORY_MASK_L |
+                                             UC_CATEGORY_MASK_M |
+                                             UC_CATEGORY_MASK_N |
+                                             UC_CATEGORY_MASK_P |
+                                             UC_CATEGORY_MASK_S)
+           && scm_c_can_put_char (port, ch))
+    /* CH is graphic and encodeable; display it.  */
+    scm_c_put_char (port, ch);
   else
+    /* CH isn't graphic or cannot be represented in PORT's encoding.  */
     {
       /* Represent CH using the character escape syntax.  */
       const char *name;
 
       name = scm_i_charname (SCM_MAKE_CHAR (ch));
       if (name != NULL)
-	scm_puts (name, port);
+        scm_puts (name, port);
+      else if (!SCM_R6RS_ESCAPES_P)
+        scm_intprint (ch, 8, port);
       else
-	PRINT_CHAR_ESCAPE (ch, port);
-    }
-}
-
-/* Write CH to PORT, escaping it if it's non-graphic or not
-   representable in PORT's encoding.  If STRING_ESCAPES_P is true and CH
-   needs to be escaped, it is escaped using the in-string escape syntax;
-   otherwise the character escape syntax is used.  */
-static void
-write_character (scm_t_wchar ch, SCM port, int string_escapes_p)
-{
-  int printed = 0;
-  scm_t_string_failed_conversion_handler strategy;
-
-  strategy = PORT_CONVERSION_HANDLER (port);
-
-  if (string_escapes_p)
-    {
-      /* Check if CH deserves special treatment.  */
-      if (ch == '"' || ch == '\\')
-	{
-	  display_character ('\\', port, iconveh_question_mark);
-	  display_character (ch, port, strategy);
-	  printed = 1;
-	}
-      else if (ch == '\n' && SCM_PRINT_ESCAPE_NEWLINES_P)
         {
-	  display_character ('\\', port, iconveh_question_mark);
-	  display_character ('n', port, strategy);
-	  printed = 1;
+          scm_puts ("x", port);
+          scm_intprint (ch, 16, port);
         }
-      else if (ch == ' ' || ch == '\n')
-	{
-	  display_character (ch, port, strategy);
-	  printed = 1;
-	}
     }
-  else
-    {
-      display_string ("#\\", 1, 2, port, iconveh_question_mark);
-
-      if (uc_combining_class (ch) != UC_CCC_NR)
-	/* Character is a combining character, so attempt to
-	   pretty-print it.  */
-	printed = write_combining_character (ch, port);
-    }
-
-  if (!printed
-      && uc_is_general_category_withtable (ch,
-					   UC_CATEGORY_MASK_L |
-					   UC_CATEGORY_MASK_M |
-					   UC_CATEGORY_MASK_N |
-					   UC_CATEGORY_MASK_P |
-					   UC_CATEGORY_MASK_S))
-    /* CH is graphic; attempt to display it.  */
-    printed = display_character (ch, port, iconveh_error);
-
-  if (!printed)
-    /* CH isn't graphic or cannot be represented in PORT's encoding.  */
-    write_character_escaped (ch, string_escapes_p, port);
-}
-
-/* Display STR to PORT from START inclusive to END exclusive.  */
-void
-scm_i_display_substring (SCM str, size_t start, size_t end, SCM port)
-{
-  int narrow_p;
-  const char *buf;
-  size_t len, printed;
-
-  buf = scm_i_string_data (str);
-  len = end - start;
-  narrow_p = scm_i_is_narrow_string (str);
-  buf += start * (narrow_p ? sizeof (char) : sizeof (scm_t_wchar));
-
-  printed = display_string (buf, narrow_p, end - start, port,
-			    PORT_CONVERSION_HANDLER (port));
-
-  if (SCM_UNLIKELY (printed < len))
-    scm_encoding_error (__func__, errno,
-			"cannot convert to output locale",
-			port, scm_c_string_ref (str, printed + start));
 }
 
 
@@ -1453,8 +1067,8 @@ scm_write (SCM obj, SCM port)
     port = scm_current_output_port ();
 
   SCM_ASSERT (scm_valid_oport_value_p (port), port, SCM_ARG2, s_write);
-
   scm_prin1 (obj, port, 1);
+
   return SCM_UNSPECIFIED;
 }
 
@@ -1468,8 +1082,8 @@ scm_display (SCM obj, SCM port)
     port = scm_current_output_port ();
 
   SCM_ASSERT (scm_valid_oport_value_p (port), port, SCM_ARG2, s_display);
-
   scm_prin1 (obj, port, 0);
+
   return SCM_UNSPECIFIED;
 }
 
@@ -1502,9 +1116,7 @@ SCM_DEFINE (scm_simple_format, "simple-format", 2, 0, 1,
   else if (scm_is_false (destination))
     {
       fReturnString = 1;
-      port = scm_mkstrport (SCM_INUM0, SCM_BOOL_F,
-			    SCM_OPN | SCM_WRTNG,
-			    FUNC_NAME);
+      port = scm_mkstrport (SCM_INUM0, SCM_BOOL_F, SCM_WRTNG, FUNC_NAME);
       destination = port;
     }
   else
@@ -1595,16 +1207,13 @@ SCM_DEFINE (scm_write_char, "write-char", 1, 1, 0,
 {
   if (SCM_UNBNDP (port))
     port = scm_current_output_port ();
+  else
+    port = SCM_COERCE_OUTPORT (port);
 
   SCM_VALIDATE_CHAR (1, chr);
-  SCM_VALIDATE_OPORT_VALUE (2, port);
+  SCM_VALIDATE_OPOUTPORT (2, port);
 
-  port = SCM_COERCE_OUTPORT (port);
-  if (!display_character (SCM_CHAR (chr), port,
-			  PORT_CONVERSION_HANDLER (port)))
-    scm_encoding_error (__func__, errno,
-			"cannot convert to output locale",
-			port, chr);
+  scm_c_put_char (port, SCM_CHAR (chr));
 
   return SCM_UNSPECIFIED;
 }
@@ -1628,7 +1237,7 @@ static int
 port_with_ps_print (SCM obj, SCM port, scm_print_state *pstate)
 {
   obj = SCM_PORT_WITH_PS_PORT (obj);
-  return scm_ptobs[SCM_PTOBNUM (obj)].print (obj, port, pstate);
+  return SCM_PORT_TYPE (obj)->print (obj, port, pstate);
 }
 
 SCM
@@ -1675,8 +1284,6 @@ scm_init_print ()
 {
   SCM type;
 
-  scm_gc_register_root (&print_state_pool);
-  scm_gc_register_root (&scm_print_state_vtable);
   type = scm_make_vtable (scm_from_locale_string (SCM_PRINT_STATE_LAYOUT),
                           SCM_BOOL_F);
   scm_set_struct_vtable_name_x (type, scm_from_latin1_symbol ("print-state"));
