@@ -181,6 +181,10 @@
         (($ $primcall 'struct-ref/immediate (struct n))
          (emit-struct-ref/immediate asm (from-sp dst) (from-sp (slot struct))
                                     (constant n)))
+        (($ $primcall 'char->integer (src))
+         (emit-char->integer asm (from-sp dst) (from-sp (slot src))))
+        (($ $primcall 'integer->char (src))
+         (emit-integer->char asm (from-sp dst) (from-sp (slot src))))
         (($ $primcall 'add/immediate (x y))
          (emit-add/immediate asm (from-sp dst) (from-sp (slot x)) (constant y)))
         (($ $primcall 'sub/immediate (x y))
