@@ -220,6 +220,7 @@
 (define *max-size-t*
   (min (+ (ash most-positive-fixnum 3) #b111)
        (1- (ash 1 48))))
+(define *max-codepoint* #x10ffff)
 
 (define-inlinable (make-unclamped-type-entry type min max)
   (vector type min max))
@@ -288,8 +289,8 @@
            (b-min (type-entry-min b)))
        (cond
         ((not (< b-min a-min)) a-min)
-        ((> 0 b-min) 0)
-        ((> &range-min b-min) &range-min)
+        ((< 0 b-min) 0)
+        ((< &range-min b-min) &range-min)
         (else -inf.0)))
      (let ((a-max (type-entry-max a))
            (b-max (type-entry-max b)))
@@ -693,8 +694,6 @@ minimum, and maximum."
 ;;; Strings.
 ;;;
 
-(define *max-char* (1- (ash 1 24)))
-
 (define-type-checker (string-ref s idx)
   (and (check-type s &string 0 *max-size-t*)
        (check-type idx &u64 0 *max-size-t*)
@@ -702,17 +701,17 @@ minimum, and maximum."
 (define-type-inferrer (string-ref s idx result)
   (restrict! s &string (1+ (&min/0 idx)) *max-size-t*)
   (restrict! idx &u64 0 (1- (&max/size s)))
-  (define! result &char 0 *max-char*))
+  (define! result &char 0 *max-codepoint*))
 
 (define-type-checker (string-set! s idx val)
   (and (check-type s &string 0 *max-size-t*)
        (check-type idx &exact-integer 0 *max-size-t*)
-       (check-type val &char 0 *max-char*)
+       (check-type val &char 0 *max-codepoint*)
        (< (&max idx) (&min s))))
 (define-type-inferrer (string-set! s idx val)
   (restrict! s &string (1+ (&min/0 idx)) *max-size-t*)
   (restrict! idx &exact-integer 0 (1- (&max/size s)))
-  (restrict! val &char 0 *max-char*))
+  (restrict! val &char 0 *max-codepoint*))
 
 (define-simple-type-checker (string-length &string))
 (define-type-inferrer (string-length s result)
@@ -743,9 +742,9 @@ minimum, and maximum."
   (define! result &flonum (&min f64) (&max f64)))
 
 (define-type-checker (scm->u64 scm)
-  (check-type scm &exact-integer 0 #xffffffffffffffff))
+  (check-type scm &exact-integer 0 &u64-max))
 (define-type-inferrer (scm->u64 scm result)
-  (restrict! scm &exact-integer 0 #xffffffffffffffff)
+  (restrict! scm &exact-integer 0 &u64-max)
   (define! result &u64 (&min/0 scm) (&max/u64 scm)))
 (define-type-aliases scm->u64 load-u64)
 
@@ -1005,9 +1004,9 @@ minimum, and maximum."
 (define-type-inferrer (uadd a b result)
   ;; Handle wraparound.
   (let ((max (+ (&max/u64 a) (&max/u64 b))))
-    (if (<= max #xffffffffffffffff)
+    (if (<= max &u64-max)
         (define! result &u64 (+ (&min/0 a) (&min/0 b)) max)
-        (define! result &u64 0 #xffffffffffffffff))))
+        (define! result &u64 0 &u64-max))))
 (define-type-aliases uadd uadd/immediate)
 
 (define-simple-type-checker (sub &number &number))
@@ -1026,7 +1025,7 @@ minimum, and maximum."
   ;; Handle wraparound.
   (let ((min (- (&min/0 a) (&max/u64 b))))
     (if (< min 0)
-        (define! result &u64 0 #xffffffffffffffff)
+        (define! result &u64 0 &u64-max)
         (define! result &u64 min (- (&max/u64 a) (&min/0 b))))))
 (define-type-aliases usub usub/immediate)
 
@@ -1078,9 +1077,9 @@ minimum, and maximum."
 (define-type-inferrer (umul a b result)
   ;; Handle wraparound.
   (let ((max (* (&max/u64 a) (&max/u64 b))))
-    (if (<= max #xffffffffffffffff)
+    (if (<= max &u64-max)
         (define! result &u64 (* (&min/0 a) (&min/0 b)) max)
-        (define! result &u64 0 #xffffffffffffffff))))
+        (define! result &u64 0 &u64-max))))
 (define-type-aliases umul umul/immediate)
 
 (define-type-checker (div a b)
@@ -1422,15 +1421,15 @@ minimum, and maximum."
   ((logior &true &false) 0 0))
 (define-type-aliases char<? char<=? char>=? char>?)
 
-(define-simple-type-checker (integer->char (&exact-integer 0 #x10ffff)))
+(define-simple-type-checker (integer->char (&u64 0 *max-codepoint*)))
 (define-type-inferrer (integer->char i result)
-  (restrict! i &exact-integer 0 #x10ffff)
-  (define! result &char (&min/0 i) (min (&max i) #x10ffff)))
+  (restrict! i &u64 0 *max-codepoint*)
+  (define! result &char (&min/0 i) (min (&max i) *max-codepoint*)))
 
 (define-simple-type-checker (char->integer &char))
 (define-type-inferrer (char->integer c result)
-  (restrict! c &char 0 #x10ffff)
-  (define! result &exact-integer (&min/0 c) (min (&max c) #x10ffff)))
+  (restrict! c &char 0 *max-codepoint*)
+  (define! result &u64 (&min/0 c) (min (&max c) *max-codepoint*)))
 
 
 
