@@ -99,7 +99,7 @@
           (let ((port (make-custom-binary-input-port "ffi-memory"
                                                      read-memory!
                                                      #f #f #f)))
-            (setvbuf port _IONBF)
+            (setvbuf port 'none)
             port)))
 
     (memory-backend dereference-word open #f)))
@@ -256,13 +256,14 @@ the matching bits, possibly with bitwise operations to extract it from BITS."
 (define %tc7-stringbuf 39)
 (define %tc7-dynamic-state 45)
 (define %tc7-frame 47)
-(define %tc7-objcode 53)
-(define %tc7-vm 55)
+(define %tc7-keyword 53)
+(define %tc7-program 69)
 (define %tc7-vm-continuation 71)
 (define %tc7-bytevector 77)
-(define %tc7-program 79)
-(define %tc7-array 85)
-(define %tc7-bitvector 87)
+(define %tc7-weak-set 85)
+(define %tc7-weak-table 87)
+(define %tc7-array 93)
+(define %tc7-bitvector 95)
 (define %tc7-port 125)
 (define %tc7-smob 127)
 
@@ -444,10 +445,9 @@ using BACKEND."
           (((_ & #x7f = %tc7-bytevector) len address)
            (let ((bv-port (memory-port backend address len)))
              (get-bytevector-n bv-port len)))
-          ((((len << 7) || %tc7-vector) weakv-data)
-           (let* ((len    (arithmetic-shift len -1))
-                  (words  (get-bytevector-n port (* len %word-size)))
-                  (vector (make-vector len)))
+          ((((len << 8) || %tc7-vector))
+           (let ((words  (get-bytevector-n port (* len %word-size)))
+                 (vector (make-vector len)))
              (visited (address -> vector)
                (fold (lambda (element index)
                        (vector-set! vector index element)
@@ -480,12 +480,14 @@ using BACKEND."
            (inferior-object 'hash-table address))
           (((_ & #x7f = %tc7-pointer) address)
            (make-pointer address))
-          (((_ & #x7f = %tc7-objcode))
-           (inferior-object 'objcode address))
-          (((_ & #x7f = %tc7-vm))
-           (inferior-object 'vm address))
+          (((_ & #x7f = %tc7-keyword) symbol)
+           (symbol->keyword (cell->object symbol backend)))
           (((_ & #x7f = %tc7-vm-continuation))
            (inferior-object 'vm-continuation address))
+          (((_ & #x7f = %tc7-weak-set))
+           (inferior-object 'weak-set address))
+          (((_ & #x7f = %tc7-weak-table))
+           (inferior-object 'weak-table address))
           (((_ & #x7f = %tc7-array))
            (inferior-object 'array address))
           (((_ & #x7f = %tc7-bitvector))
